# This file was auto-generated by salt's setup on Tuesday, 07 July 2015 @ 01:07:54 UTC.

from salt.version import SaltStackVersion

__saltstack_version__ = SaltStackVersion(2015, 8, 0, 0, 1, 0, None)
